from datetime import timedelta
from time import localtime, strftime

from pymodbus.constants import Endian
from pymodbus.payload import BinaryPayloadDecoder

from py_sma_modbus.format_unit import formatWithPrefix


class Register:
    def __init__(self, id, name, description, length, format=None, unit=None):
        self.id = id
        self.name = name
        self.description = description
        self.length = length
        self.value = None
        self.format = format
        self.unit = unit
        self.scalefactor, self.formatprecision = Register.getScale(format)
        self.noprefix = format in Register.NO_PREFIX_FORMATTING
        self._oh_name = None

    def __str__(self):
        return f"{self.id} {self.name} ({self.description}) {self.get_formattedValue()}"

    def get_formattedValue(self):
        if self.noprefix:
            return "No Value" if self.value is None else f"{self.value} {self.unit}"
        elif self.format == "Dauer":
            return (
                "No Duration"
                if self.value is None
                else str(timedelta(seconds=self.value))
            )
        elif self.format == "DT":
            return (
                "No Time"
                if self.value is None
                else strftime("%a, %d %b %Y %H:%M:%S", localtime(self.value))
            )
        else:
            return (
                "No Value"
                if self.value is None
                else formatWithPrefix(self.value, self.formatprecision, self.unit)
            )

    def get_value(self):
        return self.value

    """
    Produktübersicht SMA Solar Technology AG
    Technische Information SMA-Modbus-general-TI-de-10 14

    Format Erklärung
    Dauer Zeit in Sekunden, in Minuten oder in Stunden, je nach Modbus-Register (Ganzzahl)

    ENUM,TAGLIST Codierte Zahlenwerte. Die Aufschlüsselung der möglichen Codes fin-
    den Sie jeweils direkt unter der Bezeichnung des Modbus-Registers
    in den Zuordnungstabellen. Siehe modbuslist_de.html (Ganzzahl)

    FIX0 Dezimalzahl, kaufmännisch gerundet, ohne Nachkommastelle
    FIX1 Dezimalzahl, kaufmännisch gerundet, 1 Nachkommastelle
    FIX2 Dezimalzahl, kaufmännisch gerundet, 2 Nachkommastellen
    FIX3 Dezimalzahl, kaufmännisch gerundet, 3 Nachkommastellen
    FIX4 Dezimalzahl, kaufmännisch gerundet, 4 Nachkommastellen

    FUNKTION_SEC Das im Modbus-Register gespeicherte Datum wird bei Änderung an
    eine Funktion übergeben und startet diese. Nach Ausführen der
    Funktion ist kein Statuswert mehr gesetzt. Vor Ausführen der Funktion
    sollte in der Client-Software eine Sicherheitsabfrage vorgesehen
    werden.

    FW Firmware-Version (Ganzzahl)

    HW Hardware-Version z. B. 24 (Ganzzahl)

    IP4 4-Byte-IP-Adresse (IPv4) der Form XXX.XXX.XXX.XXX (Nicht implemetiert nur als (Ganzzahl))

    RAW Text oder Zahl. Eine RAW-Zahl hat keine Nachkommastellen und
    keine Tausender- oder sonstigen Trennzeichen. (Ganzzahl)

    REV Revisionsnummer der Form 2.3.4.5 (Ganzzahl)

    TEMP Temperaturwerte werden in speziellen Modbus-Registern in Grad
    Celsius (°C), in Grad Fahrenheit (°F) oder in Kelvin (K) gespeichert.
    Die Werte sind kaufmännisch gerundet, mit einer Nachkommastelle. (FIX1)

    TM UTC-Zeit, in Sekunden (Ganzzahl)

    UTF8 Daten im Format UTF8

    DT Datum/Uhrzeit, gemäß der Ländereinstellung (Übertragung in Se-
    kunden seit 01.01.1970) (Ganzzahl)
    """
    FORMATS = {
        "FIX1": [0.1, 1],
        "FIX2": [0.01, 2],
        "FIX3": [0.001, 3],
        "FIX4": [0.0001, 4],
        "TEMP": [0.1, 1],
    }

    NO_PREFIX_FORMATTING = {"UTF8", "TM", "REV", "RAW", "IP4", "HW", "FW"}

    OUT_OPENHAB_AS_STRING = {"UTF8", "TM", "TAGLIST", "Dauer", "DT"}

    OUT_OPENHAB_NONE_AS_0 = {"W", "A", "VAr", "VA"}

    @staticmethod
    def getScale(format):
        if format is None:
            return [1, 0]
        f = Register.FORMATS.get(format, [1, 0])
        return f

    SMA_TAGLIST = {}  # must assigned on startup! see sma.py for example


def hex_to_signed(source):
    """Convert a string hex value to a signed hexidecimal value.

    This assumes that source is the proper length, and the sign bit
    is the first bit in the first byte of the correct length.

    hex_to_signed("F") should return -1.
    hex_to_signed("0F") should return 15.
    """
    if not isinstance(source, str):
        raise ValueError("string type required")
    if 0 == len(source):
        raise ValueError("string is empty")
    sign_bit_mask = 1 << (len(source) * 4 - 1)
    other_bits_mask = sign_bit_mask - 1
    value = int(source, 16)
    return -(value & sign_bit_mask) | (value & other_bits_mask)


S16_NAN = hex_to_signed("8000")


class S16(Register):
    def __init__(self, register_id, name, description, format=None, unit=""):
        Register.__init__(self, register_id, name, description, 1, format, unit)

    def set_registers(self, registers):
        v = BinaryPayloadDecoder.fromRegisters(
            registers, byteorder=Endian.Big
        ).decode_16bit_int()
        # direct compare to 0x8000 doesn't work because 0x8000 is two's complement!!!
        # v==0x8000 is c/c++ style! This will only work in python for unsigned ints
        self.value = None if v == S16_NAN else v * self.scalefactor


S32_NAN = hex_to_signed("80000000")


class S32(Register):
    def __init__(self, register_id, name, description, format=None, unit=""):
        Register.__init__(self, register_id, name, description, 2, format, unit)

    def set_registers(self, registers):
        v = BinaryPayloadDecoder.fromRegisters(
            registers, byteorder=Endian.Big
        ).decode_32bit_int()
        self.value = None if v == S32_NAN else v * self.scalefactor


class U16(Register):
    def __init__(self, register_id, name, description, format=None, unit=""):
        Register.__init__(self, register_id, name, description, 1, format, unit)

    def set_registers(self, registers):
        v = BinaryPayloadDecoder.fromRegisters(
            registers, byteorder=Endian.Big
        ).decode_16bit_uint()
        self.value = None if v == 0xFFFF else v * self.scalefactor


class U32(Register):
    def __init__(self, register_id, name, description, format=None, unit=""):
        Register.__init__(self, register_id, name, description, 2, format, unit)

    def set_registers(self, registers):
        v = BinaryPayloadDecoder.fromRegisters(
            registers, byteorder=Endian.Big
        ).decode_32bit_uint()
        self.value = None if v == 0xFFFFFFFF or v == 0xFFFFFD else v * self.scalefactor

        if self.value and self.format == "TAGLIST":
            self.raw_value = self.value
            self.value = Register.SMA_TAGLIST.get(
                self.value, f"Unknown Value {self.value}"
            )

    def get_formattedValue(self):
        if self.value and self.format == "TAGLIST":
            return self.value
        else:
            return super().get_formattedValue()


class U64(Register):
    def __init__(self, register_id, name, description, format=None, unit=""):
        Register.__init__(self, register_id, name, description, 4, format, unit)

    def set_registers(self, registers):
        v = BinaryPayloadDecoder.fromRegisters(
            registers, byteorder=Endian.Big
        ).decode_64bit_uint()
        self.value = None if v == 0xFFFFFFFFFFFFFFFF else v * self.scalefactor


class STR32(Register):
    # STR32 Registers have variable length see modbuslist_de.html
    # Format is ignored ... it is always utf-8; unit is also ignored
    def __init__(self, register_id, name, description, length=16):
        if length < 1:
            raise "STR32 Register must have length > 0"
        Register.__init__(self, register_id, name, description, length, "UTF8", "")

    def set_registers(self, registers):
        # size is in bytes! one modbus register is 2 bytes wide
        s = BinaryPayloadDecoder.fromRegisters(
            registers, byteorder=Endian.Big
        ).decode_string(self.length * 2)
        # convert to string and remove trailing Null-Chars
        s = s.rstrip(b"\00").decode("utf-8", "ignore").strip()

        self.value = None if s == "" else s

    def get_formattedValue(self):
        return self.value



import math

SI_PREFIX = {
    -18: {"multiplier": 10**18, "prefix": "a"},
    -17: {"multiplier": 10**18, "prefix": "a"},
    -16: {"multiplier": 10**18, "prefix": "a"},
    -15: {"multiplier": 10**15, "prefix": "f"},
    -14: {"multiplier": 10**15, "prefix": "f"},
    -13: {"multiplier": 10**15, "prefix": "f"},
    -12: {"multiplier": 10**12, "prefix": "p"},
    -11: {"multiplier": 10**12, "prefix": "p"},
    -10: {"multiplier": 10**12, "prefix": "p"},
    -9: {"multiplier": 10**9, "prefix": "n"},
    -8: {"multiplier": 10**9, "prefix": "n"},
    -7: {"multiplier": 10**9, "prefix": "n"},
    -6: {"multiplier": 10**6, "prefix": "µ"},
    -5: {"multiplier": 10**6, "prefix": "µ"},
    -4: {"multiplier": 10**6, "prefix": "µ"},
    -3: {"multiplier": 10**3, "prefix": "m"},
    -2: {"multiplier": 10**3, "prefix": "m"},
    -1: {"multiplier": 10**3, "prefix": "m"},
    0: {"multiplier": 1, "prefix": ""},
    1: {"multiplier": 1, "prefix": ""},
    2: {"multiplier": 1, "prefix": ""},
    3: {"multiplier": 10**-3, "prefix": "k"},
    4: {"multiplier": 10**-3, "prefix": "k"},
    5: {"multiplier": 10**-3, "prefix": "k"},
    6: {"multiplier": 10**-6, "prefix": "M"},
    7: {"multiplier": 10**-6, "prefix": "M"},
    8: {"multiplier": 10**-6, "prefix": "M"},
    9: {"multiplier": 10**-9, "prefix": "G"},
    10: {"multiplier": 10**-9, "prefix": "G"},
    11: {"multiplier": 10**-9, "prefix": "G"},
    12: {"multiplier": 10**-12, "prefix": "T"},
    13: {"multiplier": 10**-12, "prefix": "T"},
    14: {"multiplier": 10**-12, "prefix": "T"},
    15: {"multiplier": 10**-15, "prefix": "P"},
    16: {"multiplier": 10**-15, "prefix": "P"},
    17: {"multiplier": 10**-15, "prefix": "P"},
    18: {"multiplier": 10**-18, "prefix": "E"},
}

# some units are not useful with prefix!
NO_Format = {"%", "°C", "h", "kWh", "ms", "MWh", "s", ""}


def convertNumberToNumberWithPrefix(number):
    if number == 0:
        return [number, ""]
    exponent = math.floor(math.log10(math.fabs(number)))
    exponent = max(min(exponent, 18), -18)
    return [number * SI_PREFIX[exponent]["multiplier"], SI_PREFIX[exponent]["prefix"]]


def formatWithPrefix(number, precision: int = 0, unit=""):
    if unit not in NO_Format:
        n, p = convertNumberToNumberWithPrefix(number)
    else:
        n = number
        p = ""
    return f"{n:.{precision}f} {p}{unit}"


# tests
if __name__ == "__main__":
    print(
        formatWithPrefix(1.189404e022),
        formatWithPrefix(-4.07237500000000e007),
        formatWithPrefix(1.943596e-005, 2, "F"),
        formatWithPrefix(1),
        formatWithPrefix(0.1),
        formatWithPrefix(0.001, 3, "A"),
        formatWithPrefix(0.002),
        formatWithPrefix(0.0011),
        formatWithPrefix(0.000999, 2),
        formatWithPrefix(5),
        formatWithPrefix(10),
        formatWithPrefix(100),
        formatWithPrefix(1000),
        formatWithPrefix(0),
        formatWithPrefix(-0.001),
        formatWithPrefix(-0.0011),
        formatWithPrefix(-0.000999, 1, "V"),
        formatWithPrefix(-10),
        formatWithPrefix(-100),
        formatWithPrefix(-1000),
    )



from typing import Dict, Set
from pymodbus.client.sync import ModbusTcpClient
from py_sma_modbus.registers import U32, U64, STR32, S32, Register


from py_sma_modbus.registers import Register


class Modbus:
    def __init__(self, ipAdress, ipPort=502, modbusUnit=3):
        self.registers: Set[int] = set()
        self.available_registers: Dict[int, Register] = {}
        self.unit = modbusUnit
        self._ipadress = ipAdress
        self._ipport = ipPort
        self.polling_groups = []

        self.add_tripower_register()

    # add to a set of polled registers
    def poll_register(self, register_id: int):
        if register_id not in self.available_registers:
            raise Exception(f"Register with the id {register_id} does not exist")
        else:
            self.registers.add(register_id)

    # add to the Dict of available registers
    def add_register(self, register: Register):
        self.available_registers[register.id] = register

    def run(self):
        if not self.registers:
            raise Exception("No register selected, can't poll inverter")

        self._group_register()
        return self._poll()

    def list_available_registers(self):
        for register in self.available_registers.values():
            print(register)

    def _poll(self) -> list[Register]:
        result = []

        # open socket every poll ... in case of network errors and deamon-mode the polling goes on
        with ModbusTcpClient(self._ipadress, port=self._ipport, timeout=10) as client:
            for group in self.polling_groups:
                start_id = group[0].id
                length = sum(reg.length for reg in group)

                response = client.read_holding_registers(
                    start_id, length, unit=self.unit
                )

                if not response:
                    continue

                for index, register in enumerate(group, start=0):
                    start_index = sum(register.length for register in group[0:index])
                    chunk = response.registers[
                        start_index : start_index + register.length
                    ]
                    register.set_registers(chunk)  # set and decode values

                    result.append(register)

        return result

    def _group_register(self):
        if not len(self.registers):
            return

        # self.registers is a set, all values are unique
        sorted_register = sorted(self.registers)

        polling_groups = [[]]  # new empty group

        next_id = sorted_register[0]
        for id in sorted_register:
            register = self.available_registers[id]

            if next_id != id:
                polling_groups.append(
                    []
                )  # New group, because too much distance between registers

            current_group = polling_groups[-1]  # get the Last group
            current_group.append(register)
            next_id = register.id + register.length

        self.polling_groups = polling_groups

    def add_tripower_register(self):
        """
        German registers, description and TAGLIST for SMA inverters

        Modbus® Parameter und Messwerte
        STP 15000TL-30 / STP 17000TL-30 / STP 20000TL-30 / STP 25000TL-30
        Diese Liste gilt ab folgender Firmware-Version: 2.83.03.R

        to build the list for other languages or other SMA inverters:
        1) go to https://www.sma.de/en/products/monitoring-control/modbus-protocol-interface.html
        2) under Downloads --> Background Knowledge
        3) search the "Modbus list ..." zip file for yor inverter (the zip file includes all languages)
        4) import the html-file in excel
        5) filter out on column F: "-" and "WO" ... this entries don't work .. you will get an error, if you try to read this
        6) build the register list, with a formula (STR32 has a different formula!), see the excel file in this project
        7) for the taglist:
            filter column E for taglist end enum
            extract and union all values from column O (use text editor!)
            back in excel filter unique values, split text data by ":" and "generate code" with a formula

        """
        # fmt: off
        self.add_register(U32(30001,'SMA.Modbus.Profile', 'Versionsnummer SMA Modbus-Profil', 'RAW', ''))
        self.add_register(U32(30003,'Nameplate.SusyId', 'SUSyID Modul', 'RAW', ''))
        self.add_register(U32(30005,'Nameplate.SerNum', 'Seriennummer', 'RAW', ''))
        self.add_register(U32(30051,'Nameplate.MainModel', 'Geräteklasse', 'TAGLIST', ''))
        self.add_register(U32(30053,'Nameplate.Model', 'Gerätetyp', 'TAGLIST', ''))
        self.add_register(U32(30055,'Nameplate.Vendor', 'Hersteller', 'TAGLIST', ''))
        self.add_register(U32(30057,'Nameplate.SerNum', 'Seriennummer', 'RAW', ''))
        self.add_register(U32(30059,'Nameplate.PkgRev', 'Softwarepaket', 'FW', ''))
        self.add_register(U32(30193,'DtTm.Tm', 'Systemzeit', 'DT', ''))
        self.add_register(U32(30197,'Operation.Evt.EvtNoShrt', 'Aktuelle Ereignisnummer', 'FIX0', ''))
        self.add_register(U32(30199,'Operation.RmgTms', 'Wartezeit bis Einspeisung', 'Dauer', 's'))
        self.add_register(U32(30201,'Operation.Health', 'Zustand', 'TAGLIST', ''))
        self.add_register(U32(30203,'Operation.HealthStt.Ok', 'Nennleistung im Zustand Ok', 'FIX0', 'W'))
        self.add_register(U32(30205,'Operation.HealthStt.Wrn', 'Nennleistung im Zustand Warnung', 'FIX0', 'W'))
        self.add_register(U32(30207,'Operation.HealthStt.Alm', 'Nennleistung im Zustand Fehler', 'FIX0', 'W'))
        self.add_register(U32(30211,'Operation.Evt.Prio', 'Empfohlene Aktion', 'TAGLIST', ''))
        self.add_register(U32(30213,'Operation.Evt.Msg', 'Meldung', 'TAGLIST', ''))
        self.add_register(U32(30215,'Operation.Evt.Dsc', 'Fehlerbehebungsmaßnahme', 'TAGLIST', ''))
        self.add_register(U32(30217,'Operation.GriSwStt', 'Netzrelais/-schütz', 'TAGLIST', ''))
        self.add_register(U32(30219,'Operation.DrtStt', 'Leistungsreduzierung', 'TAGLIST', ''))
        self.add_register(U32(30225,'Isolation.LeakRis', 'Isolationswiderstand', 'FIX0', 'Ohm'))
        self.add_register(U32(30231,'Inverter.WLim', 'Maximale Gerätewirkleistung', 'FIX0', 'W'))
        self.add_register(U32(30233,'Inverter.WMax', 'Eingestellte Wirkleistungsgrenze', 'FIX0', 'W'))
        self.add_register(U32(30247,'Operation.Evt.EvtNo', 'Aktuelle Ereignisnummer für Hersteller', 'FIX0', ''))
        self.add_register(U64(30513,'Metering.TotWhOut', 'Gesamtertrag', 'FIX0', 'Wh'))
        self.add_register(U64(30517,'Metering.DyWhOut', 'Tagesertrag', 'FIX0', 'Wh'))
        self.add_register(U64(30521,'Metering.TotOpTms', 'Betriebszeit', 'Dauer', 's'))
        self.add_register(U64(30525,'Metering.TotFeedTms', 'Einspeisezeit', 'Dauer', 's'))
        self.add_register(U32(30529,'Metering.TotWhOut', 'Gesamtertrag', 'FIX0', 'Wh'))
        self.add_register(U32(30531,'Metering.TotWhOut', 'Gesamtertrag', 'FIX0', 'kWh'))
        self.add_register(U32(30533,'Metering.TotWhOut', 'Gesamtertrag', 'FIX0', 'MWh'))
        self.add_register(U32(30535,'Metering.DyWhOut', 'Tagesertrag', 'FIX0', 'Wh'))
        self.add_register(U32(30537,'Metering.DyWhOut', 'Tagesertrag', 'FIX0', 'kWh'))
        self.add_register(U32(30539,'Metering.DyWhOut', 'Tagesertrag', 'FIX0', 'MWh'))
        self.add_register(U32(30541,'Metering.TotOpTms', 'Betriebszeit', 'Dauer', 's'))
        self.add_register(U32(30543,'Metering.TotFeedTms', 'Einspeisezeit', 'Dauer', 's'))
        self.add_register(U32(30559,'Operation.EvtCntUsr', 'Anzahl Ereignisse für Benutzer', 'FIX0', ''))
        self.add_register(U32(30561,'Operation.EvtCntIstl', 'Anzahl Ereignisse für Installateur', 'FIX0', ''))
        self.add_register(U32(30563,'Operation.EvtCntSvc', 'Anzahl Ereignisse für Service', 'FIX0', ''))
        self.add_register(U32(30583,'Metering.GridMs.TotWhOut', 'Zählerstand Netzeinspeise-Zähler', 'FIX0', 'Wh'))
        self.add_register(U32(30599,'Operation.GriSwCnt', 'Anzahl Netzzuschaltungen', 'FIX0', ''))
        self.add_register(S32(30769,'DcMs.Amp.MPPT1', 'DC Strom Eingang MPPT1', 'FIX3', 'A'))
        self.add_register(S32(30771,'DcMs.Vol.MPPT1', 'DC Spannung Eingang  MPPT1', 'FIX2', 'V'))
        self.add_register(S32(30773,'DcMs.Watt.MPPT1', 'DC Leistung Eingang  MPPT1', 'FIX0', 'W'))
        self.add_register(S32(30775,'GridMs.TotW', 'Leistung', 'FIX0', 'W'))
        self.add_register(S32(30777,'GridMs.W.phsA', 'Leistung L1', 'FIX0', 'W'))
        self.add_register(S32(30779,'GridMs.W.phsB', 'Leistung L2', 'FIX0', 'W'))
        self.add_register(S32(30781,'GridMs.W.phsC', 'Leistung L3', 'FIX0', 'W'))
        self.add_register(U32(30783,'GridMs.PhV.phsA', 'Netzspannung Phase L1', 'FIX2', 'V'))
        self.add_register(U32(30785,'GridMs.PhV.phsB', 'Netzspannung Phase L2', 'FIX2', 'V'))
        self.add_register(U32(30787,'GridMs.PhV.phsC', 'Netzspannung Phase L3', 'FIX2', 'V'))
        self.add_register(U32(30795,'GridMs.TotA', 'Netzstrom', 'FIX3', 'A'))
        self.add_register(U32(30803,'GridMs.Hz', 'Netzfrequenz', 'FIX2', 'Hz'))
        self.add_register(S32(30805,'GridMs.TotVAr', 'Blindleistung', 'FIX0', 'VAr'))
        self.add_register(S32(30807,'GridMs.VAr.phsA', 'Blindleistung L1', 'FIX0', 'VAr'))
        self.add_register(S32(30809,'GridMs.VAr.phsB', 'Blindleistung L2', 'FIX0', 'VAr'))
        self.add_register(S32(30811,'GridMs.VAr.phsC', 'Blindleistung L3', 'FIX0', 'VAr'))
        self.add_register(S32(30813,'GridMs.TotVA', 'Scheinleistung', 'FIX0', 'VA'))
        self.add_register(S32(30815,'GridMs.VA.phsA', 'Scheinleistung L1', 'FIX0', 'VA'))
        self.add_register(S32(30817,'GridMs.VA.phsB', 'Scheinleistung L2', 'FIX0', 'VA'))
        self.add_register(S32(30819,'GridMs.VA.phsC', 'Scheinleistung L3', 'FIX0', 'VA'))
        self.add_register(U32(30825,'Inverter.VArModCfg.VArMod', 'Betriebsart der statischen Spannungshaltung, Konfiguration der statischen Spannungshaltung', 'TAGLIST', ''))
        self.add_register(S32(30829,'Inverter.VArModCfg.VArCnstCfg.VArNom', 'Blindleistungssollwert in %', 'FIX1', '%'))
        self.add_register(S32(30831,'Inverter.VArModCfg.PFCnstCfg.PF', 'Sollwert des cos Phi, Konfiguration des cos Phi, direkte Vorgabe', 'FIX2', ''))
        self.add_register(U32(30833,'Inverter.VArModCfg.PFCnstCfg.PFExt', 'Erregungsart des cos Phi, Konfiguration des cos Phi, direkte Vorgabe', 'TAGLIST', ''))
        self.add_register(U32(30835,'Inverter.WModCfg.WMod', 'Betriebsart des Einspeisemanagements', 'TAGLIST', ''))
        self.add_register(U32(30837,'Inverter.WModCfg.WCnstCfg.W', 'Wirkleistungsbegrenzung in W', 'FIX0', 'W'))
        self.add_register(U32(30839,'Inverter.WModCfg.WCnstCfg.WNom', 'Wirkleistungsbegrenzung in %', 'FIX0', '%'))
        self.add_register(U32(30881,'Operation.PvGriConn', 'Netzanbindung der PV-Anlage ', 'TAGLIST', ''))
        self.add_register(U32(30919,'Inverter.VArModCfg.VArModDmd', 'Betriebsart der statischen Spannungshaltung bei Q on Demand, Konfiguration der statischen Spannungshaltung', 'TAGLIST', ''))
        self.add_register(U32(30925,'Spdwr.ComSocA.ConnSpd', 'Verbindungsgeschwindigkeit von SMACOM A', 'TAGLIST', ''))
        self.add_register(U32(30927,'Spdwr.ComSocA.DpxMode', 'Duplexmodus von SMACOM A', 'TAGLIST', ''))
        self.add_register(U32(30929,'Spdwr.ComSocA.Stt', 'Speedwire-Verbindungsstatus von SMACOM A', 'TAGLIST', ''))
        self.add_register(U32(30931,'Spdwr.ComSocB.ConnSpd', 'Verbindungsgeschwindigkeit von SMACOM B', 'TAGLIST', ''))
        self.add_register(U32(30933,'Spdwr.ComSocB.DpxMode', 'Duplexmodus von SMACOM B', 'TAGLIST', ''))
        self.add_register(U32(30935,'Spdwr.ComSocB.Stt', 'Speedwire-Verbindungsstatus von SMACOM B', 'TAGLIST', ''))
        self.add_register(U32(30949,'GridMs.TotPFPrc', 'Verschiebungsfaktor', 'FIX3', ''))
        self.add_register(S32(30953,'Coolsys.Cab.TmpVal', 'Innentemperatur', 'TEMP', '°C'))
        self.add_register(S32(30957,'DcMs.Amp.MPPT2', 'DC Strom Eingang MPPT2', 'FIX3', 'A'))
        self.add_register(S32(30959,'DcMs.Vol.MPPT2', 'DC Spannung Eingang MPPT2', 'FIX2', 'V'))
        self.add_register(S32(30961,'DcMs.Watt.MPPT2', 'DC Leistung Eingang  MPPT2', 'FIX0', 'W'))
        self.add_register(S32(30975,'Inverter.DclVol', 'Zwischenkreisspannung', 'FIX2', 'V'))
        self.add_register(S32(30977,'GridMs.A.phsA', 'Netzstrom Phase L1', 'FIX3', 'A'))
        self.add_register(S32(30979,'GridMs.A.phsB', 'Netzstrom Phase L2', 'FIX3', 'A'))
        self.add_register(S32(30981,'GridMs.A.phsC', 'Netzstrom Phase L3', 'FIX3', 'A'))
        self.add_register(STR32(31017,'Spdwr.ActlIp', '-', 8))
        self.add_register(STR32(31025,'Spdwr.ActlSnetMsk', '-', 8))
        self.add_register(STR32(31033,'Spdwr.ActlGwIp', '-', 8))
        self.add_register(STR32(31041,'Spdwr.ActlDnsSrvIp', '-', 8))
        self.add_register(U32(31085,'Operation.HealthStt.Ok', 'Nennleistung im Zustand Ok', 'FIX0', 'W'))
        self.add_register(S32(31159,'Operation.Dmd.VArCtl', 'Aktuelle Vorgabe Blindleistung Q', 'FIX0', 'VAr'))
        self.add_register(S32(31247,'Isolation.FltA', 'Fehlerstrom', 'FIX3', 'A'))
        self.add_register(S32(31793,'DcMs.Amp__1', 'DC Strom Eingang ?1', 'FIX3', 'A'))
        self.add_register(S32(31795,'DcMs.Am__2', 'DC Strom Eingang ?2', 'FIX3', 'A'))
        self.add_register(S32(34113,'Coolsys.Cab.TmpVal', 'Innentemperatur', 'TEMP', '°C'))
        self.add_register(U64(35377,'Operation.EvtCntUsr', 'Anzahl Ereignisse für Benutzer', 'FIX0', ''))
        self.add_register(U64(35381,'Operation.EvtCntIstl', 'Anzahl Ereignisse für Installateur', 'FIX0', ''))
        self.add_register(U64(35385,'Operation.EvtCntSvc', 'Anzahl Ereignisse für Service', 'FIX0', ''))
        self.add_register(U32(40009,'Operation.OpMod', 'Betriebszustand', 'TAGLIST', ''))
        self.add_register(U32(40013,'CntrySettings.Lang', 'Sprache der Oberfläche', 'TAGLIST', ''))
        self.add_register(U32(40029,'Operation.OpStt', 'Betriebsstatus', 'TAGLIST', ''))
        self.add_register(U32(40063,'Nameplate.CmpMain.SwRev', 'Firmware-Version des Hauptprozessors', 'FW', ''))
        self.add_register(U32(40065,'Nameplate.CmpSigProc.SwRev', 'Firmware-Version der Logikkomponente', 'FW', ''))
        self.add_register(U32(40067,'Nameplate.SerNum', 'Seriennummer', 'RAW', ''))
        self.add_register(U32(40095,'GridGuard.Cntry.VolCtl.Max', 'Spannungsüberwachung obere Maximalschwelle', 'FIX2', 'V'))
        self.add_register(U32(40109,'GridGuard.Cntry', 'Eingestellte Ländernorm', 'TAGLIST', ''))
        self.add_register(U32(40133,'GridGuard.Cntry.VRtg', 'Netz-Nennspannung', 'FIX0', 'V'))
        self.add_register(U32(40135,'GridGuard.Cntry.HzRtg', 'Nennfrequenz', 'FIX2', 'Hz'))
        self.add_register(STR32(40155,'Nameplate.MacId', '-', 2))
        self.add_register(U32(40157,'Spdwr.AutoCfgIsOn', 'Automatische Speedwire-Konfiguration eingeschaltet', 'TAGLIST', ''))
        self.add_register(STR32(40159,'Spdwr.Ip', '-', 8))
        self.add_register(STR32(40167,'Spdwr.SnetMsk', '-', 8))
        self.add_register(STR32(40175,'Spdwr.GwIp', '-', 8))
        self.add_register(U32(40185,'Inverter.VALim', 'Maximale Gerätescheinleistung', 'FIX0', 'VA'))
        self.add_register(U32(40195,'Inverter.VAMax', 'Eingestellte Scheinleistungsgrenze', 'FIX0', 'VA'))
        self.add_register(U32(40200,'Inverter.VArModCfg.VArMod', 'Betriebsart der statischen Spannungshaltung, Konfiguration der statischen Spannungshaltung', 'TAGLIST', ''))
        self.add_register(S32(40204,'Inverter.VArModCfg.VArCnstCfg.VArNom', 'Blindleistungssollwert in %', 'FIX1', '%'))
        self.add_register(S32(40206,'Inverter.VArModCfg.PFCnstCfg.PF', 'Sollwert des cos Phi, Konfiguration des cos Phi, direkte Vorgabe', 'FIX2', ''))
        self.add_register(U32(40208,'Inverter.VArModCfg.PFCnstCfg.PFExt', 'Erregungsart des cos Phi, Konfiguration des cos Phi, direkte Vorgabe', 'TAGLIST', ''))
        self.add_register(U32(40210,'Inverter.WModCfg.WMod', 'Betriebsart des Einspeisemanagements', 'TAGLIST', ''))
        self.add_register(U32(40212,'Inverter.WModCfg.WCnstCfg.W', 'Wirkleistungsbegrenzung in W', 'FIX0', 'W'))
        self.add_register(U32(40214,'Inverter.WModCfg.WCnstCfg.WNom', 'Wirkleistungsbegrenzung in %', 'FIX0', '%'))
        self.add_register(U32(40216,'Inverter.WCtlHzModCfg.WCtlHzMod', 'Betriebsart der Wirkleistungsreduktion bei Überfrequenz P(f)', 'TAGLIST', ''))
        self.add_register(U32(40218,'Inverter.WCtlHzModCfg.WCtlHzCfg.HzStr', 'Abstand der Startfrequenz zur Netzfrequenz, Konfiguration des linearen Gradienten der Momentanleistung', 'FIX2', 'Hz'))
        self.add_register(U32(40220,'Inverter.WCtlHzModCfg.WCtlHzCfg.HzStop', 'Abstand der Rücksetzfrequenz zur Netzfrequenz, Konfiguration des linearen Gradienten der Momentanleistung', 'FIX2', 'Hz'))
        self.add_register(U32(40222,'Inverter.VArModCfg.PFCtlWCfg.PFStr', 'cos Phi des Startpunktes, Konfiguration der cos Phi(P)-Kennlinie', 'FIX2', ''))
        self.add_register(U32(40224,'Inverter.VArModCfg.PFCtlWCfg.PFExtStr', 'Erregungsart des Startpunktes, Konfiguration der cos Phi(P)-Kennlinie', 'TAGLIST', ''))
        self.add_register(U32(40226,'Inverter.VArModCfg.PFCtlWCfg.PFStop', 'cos Phi des Endpunktes, Konfiguration der cos Phi(P)-Kennlinie', 'FIX2', ''))
        self.add_register(U32(40228,'Inverter.VArModCfg.PFCtlWCfg.PFExtStop', 'Erregungsart des Endpunktes, Konfiguration der cos Phi(P)-Kennlinie', 'TAGLIST', ''))
        self.add_register(U32(40230,'Inverter.VArModCfg.PFCtlWCfg.WNomStr', 'Wirkleistung des Startpunktes, Konfiguration der cos Phi(P)-Kennlinie', 'FIX0', '%'))
        self.add_register(U32(40232,'Inverter.VArModCfg.PFCtlWCfg.WNomStop', 'Wirkleistung des Endpunktes, Konfiguration der cos Phi(P)-Kennlinie', 'FIX0', '%'))
        self.add_register(U32(40234,'Inverter.WGra', 'Wirkleistungsgradient', 'FIX0', '%'))
        self.add_register(U32(40238,'Inverter.WCtlHzModCfg.WCtlHzCfg.WGra', 'Wirkleistungsgradient, Konfiguration des linearen Gradienten der Momentanleistung', 'FIX0', '%'))
        self.add_register(U32(40240,'Inverter.WCtlHzModCfg.WCtlHzCfg.HystEna', 'Aktivierung der Schleppzeigerfunktion, Konfiguration des linearen Gradienten der Momentanleistung', 'TAGLIST', ''))
        self.add_register(U32(40242,'Inverter.WCtlHzModCfg.WCtlHzCfg.HzStopWGra', 'Wirkleistungsgradient nach Rücksetzfrequenz, Konfiguration des linearen Gradienten der Momentanleistung', 'FIX0', '%'))
        self.add_register(U32(40244,'Inverter.DGSModCfg.DGSFlCfg.ArGraMod', 'Blindstromstatik, Konfiguration der vollständigen dynamischen Netzstützung', 'TAGLIST', ''))
        self.add_register(U32(40250,'Inverter.DGSModCfg.DGSMod', 'Betriebsart der dynamischen Netzstützung, Konfiguration der dynamischen Netzstützung', 'TAGLIST', ''))
        self.add_register(S32(40252,'Inverter.DGSModCfg.DGSFlCfg.DbVolNomMin', 'Untergrenze Spannungstotband, Konfiguration der vollständigen dynamischen Netzstützung', 'FIX0', '%'))
        self.add_register(U32(40254,'Inverter.DGSModCfg.DGSFlCfg.DbVolNomMax', 'Obergrenze Spannungstotband, Konfiguration der vollständigen dynamischen Netzstützung', 'FIX0', '%'))
        self.add_register(U32(40256,'Inverter.DGSModCfg.PwrCirInopVolNom', 'PWM-Sperrspannung, Konfiguration der dynamischen Netzstützung', 'FIX0', '%'))
        self.add_register(U32(40258,'Inverter.DGSModCfg.PwrCirInopTms', 'PWM-Sperrverzögerung, Konfiguration der dynamischen Netzstützung', 'FIX2', 's'))
        self.add_register(U32(40260,'Inverter.WCtlVolModCfg.CrvNum', 'Kennliniennummer, Konfiguration der Wirkleistungs-/Spannungskennlinie P(U)', 'FIX0', ''))
        self.add_register(U32(40428,'GridGuard.Cntry.FrqCtl.hhLim', 'Frequenzüberwachung mittlere Maximalschwelle', 'FIX2', 'Hz'))
        self.add_register(U32(40430,'GridGuard.Cntry.FrqCtl.hhLimTmms', 'Frequenzüberwachung mittlere Maximalschwelle Auslösezeit', 'FIX0', 'ms'))
        self.add_register(U32(40432,'GridGuard.Cntry.FrqCtl.hLim', 'Frequenzüberwachung untere Maximalschwelle', 'FIX2', 'Hz'))
        self.add_register(U32(40434,'GridGuard.Cntry.FrqCtl.hLimTmms', 'Frequenzüberwachung untere Maximalschwelle Auslösezeit', 'FIX0', 'ms'))
        self.add_register(U32(40436,'GridGuard.Cntry.FrqCtl.lLim', 'Frequenzüberwachung obere Minimalschwelle', 'FIX2', 'Hz'))
        self.add_register(U32(40438,'GridGuard.Cntry.FrqCtl.lLimTmms', 'Frequenzüberwachung obere Minimalschwelle Auslösezeit', 'FIX0', 'ms'))
        self.add_register(U32(40440,'GridGuard.Cntry.FrqCtl.llLim', 'Frequenzüberwachung mittlere Minimalschwelle', 'FIX2', 'Hz'))
        self.add_register(U32(40442,'GridGuard.Cntry.FrqCtl.llLimTmms', 'Frequenzüberwachung mittlere Minimalschwelle Auslösezeit', 'FIX0', 'ms'))
        self.add_register(U32(40446,'GridGuard.Cntry.VolCtl.MaxTmms', 'Spannungsüberwachung obere Maximalschwelle Auslösezeit', 'FIX3', 'ms'))
        self.add_register(U32(40448,'GridGuard.Cntry.VolCtl.hhLim', 'Spannungsüberwachung mittlere Maximalschwelle', 'FIX2', 'V'))
        self.add_register(U32(40450,'GridGuard.Cntry.VolCtl.hhLimTmms', 'Spannungsüberwachung mittlere Maximalschwelle Auslösezeit', 'FIX0', 'ms'))
        self.add_register(U32(40452,'GridGuard.Cntry.VolCtl.hLim', 'Spannungsüberwachung untere Maximalschwelle', 'FIX2', 'V'))
        self.add_register(U32(40456,'GridGuard.Cntry.VolCtl.hLimTmms', 'Spannungsüberwachung untere Maximalschwelle Auslösezeit', 'FIX0', 'ms'))
        self.add_register(U32(40458,'GridGuard.Cntry.VolCtl.lLim', 'Spannungsüberwachung obere Minimalschwelle', 'FIX2', 'V'))
        self.add_register(U32(40462,'GridGuard.Cntry.VolCtl.lLimTmms', 'Spannungsüberwachung obere Minimalschwelle Auslösezeit', 'FIX0', 'ms'))
        self.add_register(U32(40464,'GridGuard.Cntry.VolCtl.llLim', 'Spannungsüberwachung mittlere Minimalschwelle', 'FIX2', 'V'))
        self.add_register(U32(40466,'GridGuard.Cntry.VolCtl.llLimTmms', 'Spannungsüberwachung mittlere Minimalschwelle Auslösezeit', 'FIX0', 'ms'))
        self.add_register(U32(40472,'Inverter.PlntCtl.VRef', 'Referenzspannung, Anlagensteuerung', 'FIX0', 'V'))
        self.add_register(S32(40474,'Inverter.PlntCtl.VRefOfs', 'Referenzkorrekturspannung, Anlagensteuerung', 'FIX0', 'V'))
        self.add_register(U32(40482,'Inverter.VArGra', 'Blindleistungsgradient', 'FIX0', '%'))
        self.add_register(U32(40484,'Inverter.WGraEna', 'Aktivierung des Wirkleistungsgradienten', 'TAGLIST', ''))
        self.add_register(U32(40490,'Inverter.VArModCfg.VArCtlVolCfg.VArGraNom', 'Blindleistungsgradient, Konfiguration der Blindleistungs-/Spannungskennlinie Q(U)', 'FIX1', '%'))
        self.add_register(STR32(40497,'Nameplate.MacId', '-', 16))
        self.add_register(STR32(40513,'Spdwr.DnsSrvIp', '-', 8))
        self.add_register(U32(40575,'MltFncSw.OpMode', 'Betriebsart des Multifunktionsrelais', 'TAGLIST', ''))
        self.add_register(STR32(40631,'Nameplate.Location', '-', 16))
        self.add_register(U32(40789,'Nameplate.ComRev', 'Kommunikationsversion', 'REV', ''))
        self.add_register(U32(40809,'Nameplate.CmpSigProc.Rev', 'Umbaustand der Logikkomponente', 'FIX0', ''))
        self.add_register(U32(40915,'Inverter.WMax', 'Eingestellte Wirkleistungsgrenze', 'FIX0', 'W'))
        self.add_register(U32(40997,'Inverter.DGSModCfg.HystVolNom', 'Hysteresespannung, Konfiguration der dynamischen Netzstützung', 'FIX0', '%'))
        self.add_register(S32(41001,'Inverter.VArLimQ1', 'Maximal erreichbare Blindleistung Quadrant 1', 'FIX0', 'VAr'))
        self.add_register(S32(41007,'Inverter.VArLimQ4', 'Maximal erreichbare Blindleistung Quadrant 4', 'FIX0', 'VAr'))
        self.add_register(S32(41009,'Inverter.PFLimQ1', 'Minimal erreichbarer cos(Phi) Quadrant 1', 'FIX3', ''))
        self.add_register(S32(41015,'Inverter.PFLimQ4', 'Minimal erreichbarer cos(Phi) Quadrant 4', 'FIX3', ''))
        self.add_register(U32(41017,'Inverter.UtilCrvCfg.Crv.CrvTms', 'Einstellzeit des Kennlinienarbeitspunktes, Konf. der Netzintegrationskennlinie 1', 'FIX1', 's'))
        self.add_register(U32(41019,'Inverter.UtilCrvCfg.Crv.RmpDec', 'Absenkungsrampe, Konfiguration der Netzintegrationskennlinie 1', 'FIX1', '%'))
        self.add_register(U32(41021,'Inverter.UtilCrvCfg.Crv.RmpInc', 'Steigerungsrampe, Konfiguration der Netzintegrationskennlinie 1', 'FIX1', '%'))
        self.add_register(U32(41023,'Inverter.UtilCrvCfg.Crv.NumPt', 'Anzahl zu verwendender Punkte, Konfiguration der Netzintegrationskennlinie 1', 'FIX0', ''))
        self.add_register(U32(41025,'Inverter.UtilCrvCfg.Crv.XRef', 'X-Achsen-Referenz, Konf. der Netzintegrationskennlinie 1', 'TAGLIST', ''))
        self.add_register(U32(41027,'Inverter.UtilCrvCfg.Crv.YRef', 'Y-Achsen-Referenz, Konf. der Netzintegrationskennlinie 1', 'TAGLIST', ''))
        self.add_register(S32(41029,'Inverter.UtilCrvCfg.Crv.XVal1', 'X-Wert 1, Konfiguration der Netzintegrationskennlinie 1', 'FIX3', ''))
        self.add_register(S32(41031,'Inverter.UtilCrvCfg.Crv.YVal1', 'Y-Wert 1, Konfiguration der Netzintegrationskennlinie 1', 'FIX3', ''))
        self.add_register(S32(41033,'Inverter.UtilCrvCfg.Crv.XVal2', 'X-Wert 2, Konfiguration der Netzintegrationskennlinie 1', 'FIX3', ''))
        self.add_register(S32(41035,'Inverter.UtilCrvCfg.Crv.YVal2', 'Y-Wert 2, Konfiguration der Netzintegrationskennlinie 1', 'FIX3', ''))
        self.add_register(S32(41037,'Inverter.UtilCrvCfg.Crv.XVal3', 'X-Wert 3, Konfiguration der Netzintegrationskennlinie 1', 'FIX3', ''))
        self.add_register(S32(41039,'Inverter.UtilCrvCfg.Crv.YVal3', 'Y-Wert 3, Konfiguration der Netzintegrationskennlinie 1', 'FIX3', ''))
        self.add_register(S32(41041,'Inverter.UtilCrvCfg.Crv.XVal4', 'X-Wert 4, Konfiguration der Netzintegrationskennlinie 1', 'FIX3', ''))
        self.add_register(S32(41043,'Inverter.UtilCrvCfg.Crv.YVal4', 'Y-Wert 4, Konfiguration der Netzintegrationskennlinie 1', 'FIX3', ''))
        self.add_register(S32(41045,'Inverter.UtilCrvCfg.Crv.XVal5', 'X-Wert 5, Konfiguration der Netzintegrationskennlinie 1', 'FIX3', ''))
        self.add_register(S32(41047,'Inverter.UtilCrvCfg.Crv.YVal5', 'Y-Wert 5, Konfiguration der Netzintegrationskennlinie 1', 'FIX3', ''))
        self.add_register(S32(41049,'Inverter.UtilCrvCfg.Crv.XVal6', 'X-Wert 6, Konfiguration der Netzintegrationskennlinie 1', 'FIX3', ''))
        self.add_register(S32(41051,'Inverter.UtilCrvCfg.Crv.YVal6', 'Y-Wert 6, Konfiguration der Netzintegrationskennlinie 1', 'FIX3', ''))
        self.add_register(S32(41053,'Inverter.UtilCrvCfg.Crv.XVal7', 'X-Wert 7, Konfiguration der Netzintegrationskennlinie 1', 'FIX3', ''))
        self.add_register(S32(41055,'Inverter.UtilCrvCfg.Crv.YVal7', 'Y-Wert 7, Konfiguration der Netzintegrationskennlinie 1', 'FIX3', ''))
        self.add_register(S32(41057,'Inverter.UtilCrvCfg.Crv.XVal8', 'X-Wert 8, Konfiguration der Netzintegrationskennlinie 1', 'FIX3', ''))
        self.add_register(S32(41059,'Inverter.UtilCrvCfg.Crv.YVal8', 'Y-Wert 8, Konfiguration der Netzintegrationskennlinie 1', 'FIX3', ''))
        self.add_register(U32(41065,'Inverter.UtilCrvCfg.Crv2.CrvTms', 'Einstellzeit des Kennlinienarbeitspunktes, Konfiguration der Netzintegrationskennlinie 2', 'FIX1', 's'))
        self.add_register(U32(41067,'Inverter.UtilCrvCfg.Crv2.RmpDec', 'Absenkungsrampe, Konfiguration der Netzintegrationskennlinie 2', 'FIX1', '%'))
        self.add_register(U32(41069,'Inverter.UtilCrvCfg.Crv2.RmpInc', 'Steigerungsrampe, Konfiguration der Netzintegrationskennlinie 2', 'FIX1', '%'))
        self.add_register(U32(41071,'Inverter.UtilCrvCfg.Crv2.NumPt', 'Anzahl zu verwendender Punkte, Konfiguration der Netzintegrationskennlinie 2', 'FIX0', ''))
        self.add_register(U32(41073,'Inverter.UtilCrvCfg.Crv2.XRef', 'Eingangseinheit, Konfiguration der Netzintegrationskennlinie 2', 'TAGLIST', ''))
        self.add_register(U32(41075,'Inverter.UtilCrvCfg.Crv2.YRef', 'Ausgangsreferenz, Konfiguration der Netzintegrationskennlinie 2', 'TAGLIST', ''))
        self.add_register(S32(41077,'Inverter.UtilCrvCfg.Crv2.XVal1', 'X-Wert 1, Konfiguration der Netzintegrationskennlinie 2', 'FIX3', ''))
        self.add_register(S32(41079,'Inverter.UtilCrvCfg.Crv2.YVal1', 'Y-Wert 1, Konfiguration der Netzintegrationskennlinie 2', 'FIX3', ''))
        self.add_register(S32(41081,'Inverter.UtilCrvCfg.Crv2.XVal2', 'X-Wert 2, Konfiguration der Netzintegrationskennlinie 2', 'FIX3', ''))
        self.add_register(S32(41083,'Inverter.UtilCrvCfg.Crv2.YVal2', 'Y-Wert 2, Konfiguration der Netzintegrationskennlinie 2', 'FIX3', ''))
        self.add_register(S32(41085,'Inverter.UtilCrvCfg.Crv2.XVal3', 'X-Wert 3, Konfiguration der Netzintegrationskennlinie 2', 'FIX3', ''))
        self.add_register(S32(41087,'Inverter.UtilCrvCfg.Crv2.YVal3', 'Y-Wert 3, Konfiguration der Netzintegrationskennlinie 2', 'FIX3', ''))
        self.add_register(S32(41089,'Inverter.UtilCrvCfg.Crv2.XVal4', 'X-Wert 4, Konfiguration der Netzintegrationskennlinie 2', 'FIX3', ''))
        self.add_register(S32(41091,'Inverter.UtilCrvCfg.Crv2.YVal4', 'Y-Wert 4, Konfiguration der Netzintegrationskennlinie 2', 'FIX3', ''))
        self.add_register(S32(41093,'Inverter.UtilCrvCfg.Crv2.XVal5', 'X-Wert 5, Konfiguration der Netzintegrationskennlinie 2', 'FIX3', ''))
        self.add_register(S32(41095,'Inverter.UtilCrvCfg.Crv2.YVal5', 'Y-Wert 5, Konfiguration der Netzintegrationskennlinie 2', 'FIX3', ''))
        self.add_register(S32(41097,'Inverter.UtilCrvCfg.Crv2.XVal6', 'X-Wert 6, Konfiguration der Netzintegrationskennlinie 2', 'FIX3', ''))
        self.add_register(S32(41099,'Inverter.UtilCrvCfg.Crv2.YVal6', 'Y-Wert 6, Konfiguration der Netzintegrationskennlinie 2', 'FIX3', ''))
        self.add_register(S32(41101,'Inverter.UtilCrvCfg.Crv2.XVal7', 'X-Wert 7, Konfiguration der Netzintegrationskennlinie 2', 'FIX3', ''))
        self.add_register(S32(41103,'Inverter.UtilCrvCfg.Crv2.YVal7', 'Y-Wert 7, Konfiguration der Netzintegrationskennlinie 2', 'FIX3', ''))
        self.add_register(S32(41105,'Inverter.UtilCrvCfg.Crv2.XVal8', 'X-Wert 8, Konfiguration der Netzintegrationskennlinie 2', 'FIX3', ''))
        self.add_register(S32(41107,'Inverter.UtilCrvCfg.Crv2.YVal8', 'Y-Wert 8, Konfiguration der Netzintegrationskennlinie 2', 'FIX3', ''))
        self.add_register(U32(41111,'GridGuard.Cntry.VolCtl.MinEff', 'Spannungsüberwachung untere Minimalschwelle als Effektivwert', 'FIX2', 'V'))
        self.add_register(U32(41113,'GridGuard.Cntry.VolCtl.MinEffTmms', 'Spannungsüberwachung untere Minimalschwelle als Effektivwert Auslösezeit', 'FIX0', 'ms'))
        self.add_register(U32(41115,'GridGuard.Cntry.VolCtl.MaxEff', 'Spannungsüberwachung obere Maximalschwelle als Effektivwert', 'FIX2', 'V'))
        self.add_register(U32(41117,'GridGuard.Cntry.VolCtl.MaxEffTmms', 'Spannungsüberwachung obere Maximalschwelle als Effektivwert Auslösezeit', 'FIX0', 'ms'))
        self.add_register(U32(41121,'GridGuard.CntrySet', 'Setze Ländernorm', 'FUNKTION_SEC', ''))
        self.add_register(U32(41123,'GridGuard.Cntry.VolCtl.ReconMin', 'Min. Spannung zur Wiederzuschaltung', 'FIX2', 'V'))
        self.add_register(U32(41125,'GridGuard.Cntry.VolCtl.ReconMax', 'Max. Spannung zur Wiederzuschaltung', 'FIX2', 'V'))
        self.add_register(U32(41127,'GridGuard.Cntry.FrqCtl.ReconMin', 'Untere Frequenz für Wiederzuschaltung', 'FIX2', 'Hz'))
        self.add_register(U32(41129,'GridGuard.Cntry.FrqCtl.ReconMax', 'Obere Frequenz für Wiederzuschaltung', 'FIX2', 'Hz'))
        self.add_register(U32(41131,'DcCfg.StrVol', 'minimale Spannung Eingang ', 'FIX2', 'V'))
        self.add_register(U32(41133,'DcCfg.StrVol', 'minimale Spannung Eingang ', 'FIX2', 'V'))
        self.add_register(U32(41155,'DcCfg.StrTms', 'Startverzögerung Eingang ', 'FIX0', 's'))
        self.add_register(U32(41157,'DcCfg.StrTms', 'Startverzögerung Eingang ', 'FIX0', 's'))
        self.add_register(U32(41169,'GridGuard.Cntry.LeakRisMin', 'Minimaler Isolationswiderstand', 'FIX0', 'Ohm'))
        self.add_register(U32(41171,'Metering.TotkWhOutSet', 'Setze Gesamtertrag', 'FIX0', 'kWh'))
        self.add_register(U32(41173,'Metering.TotOpTmhSet', 'Setze Gesamte Betriebszeit am Netzanschlusspunkt', 'Dauer', 'h'))
        self.add_register(U32(41193,'Inverter.CtlComCfg.WCtlCom.CtlComMssMod', 'Betriebsart für ausbleibende Wirkleistungsbegrenzung', 'TAGLIST', ''))
        self.add_register(U32(41195,'Inverter.CtlComCfg.WCtlCom.TmsOut', 'Timeout für ausbleibende Wirkleistungsbegrenzung', 'Dauer', 's'))
        self.add_register(U32(41197,'Inverter.CtlComCfg.WCtlCom.FlbWNom', 'Fallback Wirkleistungsbegrenzung P in % von WMax für ausbleibende Wirkleistungsbegrenzung', 'FIX2', '%'))
        self.add_register(U32(41219,'Inverter.CtlComCfg.VArCtlCom.CtlComMssMod', 'Betriebsart für ausbleibende Blindleistungsregelung', 'TAGLIST', ''))
        self.add_register(U32(41221,'Inverter.CtlComCfg.VArCtlCom.TmsOut', 'Timeout für ausbleibende Blindleistungsregelung', 'Dauer', 's'))
        self.add_register(S32(41223,'Inverter.CtlComCfg.VArCtlCom.FlbVArNom', 'Fallback Blindleistung Q in % von WMax für ausbleibende Blindleistungsregelung', 'FIX2', '%'))
        self.add_register(U32(41225,'Inverter.CtlComCfg.PFCtlCom.CtlComMssMod', 'Betriebsart für ausbleibende cos Phi-Vorgabe', 'TAGLIST', ''))
        self.add_register(U32(41227,'Inverter.CtlComCfg.PFCtlCom.TmsOut', 'Timeout für ausbleibende cos Phi-Vorgabe', 'Dauer', 's'))
        self.add_register(S32(41229,'Inverter.CtlComCfg.PFCtlCom.FlbPF', 'Fallback cos Phi für ausbleibende cos Phi-Vorgabe', 'FIX4', ''))
        self.add_register(U64(42109,'Modbus.UnitID', 'Unit ID des Wechselrichters', 'RAW', ''))
        self.add_register(U32(43090,'SMA.GridGuard.Code', 'SMA Grid Guard-Code', 'RAW', ''))
        # fmt: on
